// ================== Morse Code Practice & Exercise applets =================
// 2020-2023 Ron Miller (RonM9 on Instructables.com) the code depends on utilities in my STEM Platform library 
// This code set evolved over a few years as seemed fit in moments, and has not been cleared up.
// This code serves well for learning Morse Code, not so much for learning Computer Science.
// license: Attribution-NonCommercial-ShareAlike

// =================== local Function prototype predefinitions, for reference & navigation
void MC_freeForm(); // -------- allow free form Morse code practice until btn4 pressed
void A2Z_Morse_Exercise(); // Exercise keying in A-Z in order. or alphabet phonetic words Apple-Zebra or NATO Phonetics
void Morse_AB_Prompt(); // Follow random letter with the MC for the letter following it alphabetically
void MC_Relay_exercise(); // Write down the letters of a 'word' as you hear them. Then try to key them in (as if to relay)
void MC_Receive_Practice(); // Practice listening to and making out MC characters at your own pace.
void Tx_Line_test(); // Practice Transmitting, Key in complete lines as seen on the Monitor screen.

// -----------------------------------------    utilities used herein
void MC_setup(); // Morse Code setup
bool isPuntc(byte charCode); // identifies punctuation as we can be tolerant of delays keying them in (added 0-9 as well)
void isPuntc_debugTest();
void dit_dah(unsigned short int delayms); // Used for ditdah audio while user has key pressed
void ditAudio(unsigned short int delayms); // over duty cycle adjusted for freq.
void dahAudio(unsigned short int delayms); // lower freq ( w/ longer ON duty cycle for Piezo)
void playMorseCode(byte ditdahs, int pace);
void playStr(char *wrd, int pace);        // ment for message announcements
void reportRate();
byte fetchMorseChar(bool echo);
byte fetchMorse_1Key(bool echo); // get one char (ditdah set) using single button keying (straight CW key)
byte fetchMorse_2Key(bool echo); // using Two button keying (or a dual lever paddle)
byte fetchMorse_22Key(bool echo); // using Two button keying (or a dual lever paddle)
char dida2Ascii(byte ditdahs);
byte Ascii2dida(char Achar);
void set_didaStr(byte ditdahs, char ddStr[]); // returns string representation in global var: 'didaStr'
int  pickOne(int min, int max); // returns value in range min - max-1


// ---------- Morse code support data structure ------------
int numWrds = 900; //or 500
#define IsPunct  (cinput>32 && !(toupper(cinput)>64 && toupper(cinput)<91)
bool prtChr, pAdj;

unsigned int avgPeriod, pAveraging;
unsigned int avgDit=800,avgDah=2400;
unsigned int ditPeriod;
int farns, deFF;  // Farnsworth Factor, multiplier for std to Farnsworth inter char & word periods
bool pOpen = true; // expect an open ( else next should be close )

bool oneKey = true; // straight single contact key   (TBD: false=using a 2 contact paddle)
//bool batOp=false; // would-be indicator for battery Operation (w/o serial monitor)
char didaStr[10];

const char alpha[] = "ZABCDEFGHIJKLMNOPQRSTUVWXYZA";
const char nonAlpha[] = " 1234567890.,?/'-:\"+!=@$)&\n(_~";
//       1    2    3    4    5    6    7    8    9    0
const byte nonAlpha_dida[] = {0, 0x10, 0x18, 0x1C, 0x1E, 0x1F, 0xEF, 0xE7, 0xE3, 0xE1, 0xE0,
                              0x2A, 0xCC, 0x33, 0xED, 0x21, 0xDE, 0xC7, 0x2D, 0x15, 0xD4, 0xEE, 0x25, 0x76, 0xD2, 0x17, 0x0A, 0xE9, 0x0};
                            //  .    ,     ?     /     '     -     :     "     +     !     =     @     $     )     &     \n    (     _

void MC_setup() {   // Morse Code setup
  //  ditPeriod= (60000)/(240*2);  // in msecs, for ~10WPM to start off  (240 in 60,000 ms is 5WPM)
  ditPeriod = (60000) / (240 * 2 + 120); // in msecs, for ~12WPM to start off  (240 in 60,000 ms is 5WPM)
  //  ditPeriod= (60000)/(240*3);      // in msecs, for ~15WPM to start off  (240 in 60,000 ms is 5WPM)
  //  ditPeriod= (60000)/(240*4);      // in msecs, for ~20WPM to start off  (240 in 60,000 ms is 5WPM)

  // Farnsworth calculations
  int cWPM = 18, sWPM = 10;
  float fp = 1000. * (60 / sWPM - 37.2 / cWPM) / 19.; // inter time period of Farnsworth timing
  farns = fp / (1000 * 1.2 / cWPM); // factor to convert std to Farnsworth unit time period
  deFF = farns;

  avgPeriod = 2 * ditPeriod;
  pAveraging = 4 * avgPeriod;

  if (btn1) {
    farns = 1; // disable Farnsworth effect if btn1 held down at power up.
    Serial.println("Standard inter-char & word timing used.");
  } else {
    Serial.print("Farnsworth ratio: "); Serial.print(sWPM); Serial.print("/"); Serial.println(cWPM);
  }

  //  Serial.print(" Keying rate (in WPM): "); Serial.print((1000.0*60.0) / (240.0*ditPeriod/5.0));
  //  Serial.print("  + Farnsworth spacing: x"); Serial.println(farns);

}

// ================== Morse Code Practice & Exercise applets =================
// -----------------------
#include <avr/pgmspace.h>
char buf[30];

//  words either sized 500 or 900 (now only 825)
//const PROGMEM char wordStr[] = "    ,Quick,brown,fox,jumps,over,the,lazy,dog,Pack,my ,box,with,five,dozen,liquor,jugs,Cozy,lynx,gives,smart,squid,asks,for,pen,Hive,of ,boxing,wizards,jump,quickly,Garden,Gone,Wild,But,that's,too,okay,very,been,guess,time,going,into,those,did,work,other,I've,even,our,any,yes,his,had,were,where,your,she,which,will,say,who,make,man,than,could,state,only,new,year,come,take,these,use,first,may,much,give,over,most,find,day,also,after,way,many,must,look,before,great,back,long,down,should,through,own,two,here,last,want,didn't,something,being,still,put,never,made,went,thought,again,off,why,can't,quite,actually,might,world,life,another,used,night,old,same,always,next,came,doing,week,house,none,each,feel,seem,high,place,little,nation,hand,tell,write,become,show,both,between,need,call,develop,under,move,general,school,begin,while,part,turn,real,leave,point,form,child,few,small,since,against,ask,late,interest,large,person,end,open,public,follow,during,present,without,hold,govern,around,possible,head,consider,word,program,problem,however,lead,system,set,order,eye,plan,run,keep,face,fact,group,play,stand,increase,early,course,change,help,line,send,listen,sound,audio,quality,xmitt,receive,morse,code,electronic,class,STEM,arduino,kids,email,text,phone,in-box,o'clock,wifi,video,games,arm,leg,foot,LOL!,?Who?,(408),34-26-48,KI7NEW,(3+1=4),3.1416,CQCQ,8/27/49,6:15am,Junk,jumper,jewlery,jam,pajamas,adjust,equal,quote,vote,voltage,box,fix,max,tax,six,taxes,zero,lazy,zone,crazy,Amazon,Z-axis,home,page,has,search,free,i10n,site,news,contact,business,web,view,online,click,its,service,price,date,top,list,name,health,music,buy,data,product,post,her,city,add,policy,number,such,please,available,support,message,best,software,info,rights,books,links,review,years,privacy,book,items,company,read,user,said,research,university,mail,full,map,reviews,days,management,united,hotel,item,center,eBay,store,travel,comments,development,report,member,details,terms,hotels,type,local,using,results,office,education,national,car,design,posted,internet,address,community,within,states,area,dvd,shipping,reserved,subject,forum,family,based,black,check,special,prices,website,index,women,sign,file,link,today,tech.,south,case,project,pages,version,section,found,sports,related,security,county,American,photo,game,members,power,care,network,computer,three,total,following,download,him,per,access,north,resource,current,posts,big,media,law,control,water,history,pictures,size,art,personal,including,guide,shop,directory,board,location,white,rating,rate,government,children,USA,return,students,shopping,account,times,sites,level,digital,profile,previous,events,love,john,main,hours,image,department,title,descripe,non,insurance,shall,property,money,every,listing,content,country,private,visit,save,tools,low,reply,customer,compare,movies,include,college,value,article,york,card,jobs,provide,food,source,author,different,press,learn,sale,print,job,Canada,process,teen,room,stock,training,credit,join,science,men,category,advanced,west,sales,english,left,team,estate,condition,select,windows,photos,party,";
const PROGMEM char wordStr[] = "    ,that's,too,okay,very,been,guess,time,going,into,those,did,work,other,I've,even,our,any,yes,his,had,were,where,your,she,which,will,say,who,make,man,than,could,state,only,new,year,come,take,these,use,first,may,much,give,over,most,find,day,also,after,way,many,must,look,before,great,back,long,down,should,through,own,two,here,last,want,didn't,something,being,still,put,never,made,went,thought,again,off,why,can't,quite,actually,might,world,life,another,used,night,old,same,always,next,came,doing,week,house,none,each,feel,seem,high,place,little,nation,hand,tell,write,become,show,both,between,need,call,develop,under,move,school,begin,while,part,turn,real,leave,point,form,child,few,small,since,against,ask,late,interest,large,person,end,open,public,follow,during,present,without,hold,govern,around,possible,head,consider,word,program,problem,however,lead,system,set,order,eye,plan,run,keep,face,fact,group,play,stand,increase,early,course,change,help,line,send,listen,sound,audio,quality,xmitt,receive,code,electronic,tech.,class,STEM,arduino,kids,email,text,phone,in-box,o'clock,wifi,video,games,arm,leg,foot,LOL!,?Who?,(408),34-26-48,KI7NEW,(3+1=4),3.1416,CQCQ,8/27/49,6:15am,Junk,jumper,jewlery,jam,pajamas,adjust,equal,quote,vote,voltage,box,fix,max,tax,six,taxes,zero,lazy,zone,crazy,Amazon,Z-axis,BTW,R=roger,TU:ThankYou,RTU:R+TU,TNX:thanks,QRA:name?,AGN=again,ANT=antenna,PSE=please,GA:go-ahead,HI:laughter,NR:number,OM:old-man,PWR=power,WX:weather,QRL: BUSY,QRN: NOISE,QRQ (FASTER),QRS (SLOWER),QTH LOCATION,QSB: FADING,QSY CHG FREQ,QRP low Pwr,home,page,has,search,free,i10n,site,news,contact,business,web,view,online,click,its,service,price,date,top,list,name,health,music,buy,data,product,post,her,city,add,policy,number,such,please,available,support,message,best,software,info,rights,books,links,review,years,privacy,book,items,company,read,user,said,research,university,mail,full,map,reviews,days,management,united,hotel,item,center,eBay,store,travel,comments,development,report,member,details,terms,hotels,type,local,using,results,office,education,national,car,design,posted,internet,address,community,within,states,area,dvd,shipping,reserved,subject,forum,family,based,black,check,special,prices,website,index,women,sign,file,link,today,south,case,project,pages,version,section,found,sports,related,security,county,American,photo,game,members,power,care,network,computer,three,total,following,download,him,per,access,north,resource,current,posts,big,media,law,control,water,history,pictures,size,art,personal,including,guide,shop,directory,board,location,white,rating,rate,government,children,USA,return,students,shopping,account,times,sites,level,digital,profile,previous,events,love,john,main,hours,image,department,title,descripe,non,insurance,shall,property,money,every,listing,content,country,private,visit,save,tools,low,reply,customer,compare,movies,include,college,value,article,york,card,jobs,provide,food,source,author,different,press,learn,sale,print,job,Canada,process,teen,room,stock,training,credit,join,science,men,category,advanced,west,sales,english,left,team,estate,condition,select,windows,photos,party,thread,note,live,gallery,table,register,market,library,action,start,series,model,features,air,industry,human,provided,required,second,hot,accessory,cost,movie,forums,better,question,yahoo,medical,Army,Navy,doctor,nurse,police,fireman,bit,byte,road,months,sort,moment,street,town,Mexico,door,side,mother,Jan.,Feb.,March,April,May,June,Jul.,Aug.,Sept.,Oct.,Nov.,Dec.,Monday,Tue.,Wed.,Thur.,Fri.,Saturday,Sunday,Europe,Asia,Africa,Mars,HAM,Balun,dipole,end feed,OCF,long wire,J-pole,half-wave,rag chew,Yaesu,Baofeng,Kenwood,ICOM,PTT,band,Chirp,SDR,FT-8,FCC,amplifier,tuner,SSB,LSB,Walkie-Talkie,80M,40M,VHF,UHF,Morse,Novice,General,Extra,license,Net,SWR,coax,cable,rope,leader,ladder,12V,USB,Blue Tooth,Laptop,cell ph.,coil,caps,xister,LEDs,ICs,filter,oscillate,impedance,50Ohm,SO239,PL259,BNC,SMA,RG8x,RG174,resistance,solder,circuit,bought,sold,sell,built,DIY,kits,measure,GRM,FRS,Solar,wait,field,bench,break,bend,emit,radiate,BFO,AGC,AFC,VCO,comm,package,bank,chock,switch,flash,portable,friend,LiFePo4,H2O,CO2,inverter,UPS,panel,shack,voice,analog,rig,beam,yrs,temp,loop,Zepp,Yagi,G5RV,quad,speaker,microphone,Heath,Windom,Doublet,sloper,horizontal,PSK31,MFJ,599,EndFed,Inv-V,vertical,roof,tree,snack,kilogram,pound,meter,inch,feet,seconds,minutes,sleep,drive,lunch,drink,beverage,amateur,hail,snow,rain,cold,warm,cool,clear,sunny,windy,cloudy,storms,overcast,freezing,dark,station,channel,operator,Amplitude,modulation,frequency,band-pass,beacon,chassis,repeater,conduce,iron-core,crystal,detector,element,dummy,earth,emergency,emission,energy,signal,noise,overload,fuse,harmonic,inductor,insulator,ionize,lightning,metric,micro-,milli-,primary,traffic,propagation,RTTY,reflection,resistor,safety,skip,sky wave,sunspots,cycle,third-party,transmission,variable,ARES,balanced,trap,shield,equipment,azimuth,bandwidth,Polarity,carrier,components,breaker,controlled,connector,ferrite,toroid,procedure,exam,cutoff,privilege,spectrum,manual,error,feed line,display,simplex,coverage,plane,identify,permit,EMI,keyboard,keyer,technician,logbook,loose,magnetic,matching,maximum,military,negative,positive,node,operating,packet,scatter,radio,sending,speed,reflected,schematic,thermal,density,rotor,satellite,shunt,sine wave,device,speech,splatter,stable,straight,stray,terminal,tower,tubes,tuned,length,width,height,club,";
              // 2nd line has ~800 entries (of 3-11 letters)
              // needs wrd buffer at least 2*max distance between ','s  so ~26 (for max word len=12+',' *2) for 2nd line above.  And ~22 for 1st line above 

// ================== Local Morse Code Support Utility Functions =================

// -----------------------
bool isPuntc(byte charCode) { // identifies punctuation as we can be tolerant of delays keying them in (added 0-9 as well)
  //  return (charCode>32 && !(toupper(charCode)>64 && toupper(charCode)<91) && !(toupper(charCode)>47 && toupper(charCode)<58));
  return (charCode > 32 && !(toupper(charCode) > 64 && toupper(charCode) < 91));
}

void isPuntc_debugTest() {
  /***  char c;
    Serial.println(" Char,  isPuntc(c),  dida2Ascii(Ascii2dida(c))");
    for (byte b=32; b<=127; b++) {c=b; Serial.print(c);}
    Serial.println();
    for (byte b=32; b<=127; b++) { Serial.print(isPuntc(b));}
    Serial.println();
    Serial.print(' '); // dida2Ascii(Ascii2dida(' '))=0   so print a ' ' holder for it
    for (byte b=33; b<=127; b++) Serial.print(dida2Ascii(Ascii2dida(b)));  Serial.println();
  ***/

  // List/Test all of the words in our vocabulary
  int wi, wlen ;
  char wrd[30], tmpw[30];
  char *cptr;
  int strInx;
  int deltInx;

  strInx = 0;
  for (wi = 0; wi < 900; wi++) {
    //strcpy_P(wrd, (char *)pgm_read_word(&(words[wi])));
    do {  //    randomly select a 'word'
      //        strInx = random(0,strlen(wordStr)-8); // randomly select an area (// out for listing in order)
      strncpy_P(wrd, (char *) & (wordStr[strInx]), 26 );
      wrd[26] = 0;
      strcpy(tmpw, strchr(wrd, ',') + 1); // pull out a word
      deltInx = strlen(wrd) - strlen(tmpw);
      cptr = strchr(tmpw, ',');
      if (cptr == NULL) continue;
      wlen = (int) (cptr - tmpw);
      strncpy(wrd, tmpw, wlen);
      wrd[wlen] = 0;
      //        Serial.print("  wlen is "); Serial.print(wlen);  Serial.print(" : "); Serial.println(wrd);
      //        Serial.print(" strInx is "); Serial.print(strInx);  Serial.print(" : "); Serial.println(wrd);
    } while (wlen < 1);
    strInx = strInx + deltInx;
    if (strInx > (strlen(wordStr) - 4)) return;

    wlen = strlen(wrd);
    Serial.print(wi); Serial.print("\t"); Serial.print(wlen);
    Serial.print("\t*"); Serial.print(wrd); Serial.println("*");
  }
}


// -----------------------
void dit_dah(unsigned short int delayms) { // Used for ditdah audio while user has key pressed
                                          // higher freq & longer duty cycle for piezo
  delayms = delayms - (delayms / 5);
  for (int i = 0; i < delayms; i++) {
    digitalWrite ( AUDIO_OUT, HIGH);
    if (Loud)  digitalWrite ( AUDIO_NEG, LOW);
    delayMicroseconds(840);   // 750 840
    digitalWrite ( AUDIO_OUT, LOW);
    if (Loud)  digitalWrite ( AUDIO_NEG, HIGH);
    delayMicroseconds(560);   // 500 560
  }
  //digitalWrite ( AUDIO_OUT, LOW);;
  if (Loud)  digitalWrite ( AUDIO_NEG, LOW);
}

// -----------------------
void ditAudio(unsigned short int delayms) { //  over duty cycle adjusted for freq.
  // ON duration extra needed for good piezo sound
  //  delayms = delayms + (delayms/2);  // for use with 400/266
  delayms = delayms - (delayms / 5); // for use with 700/550
  for (int i = 0; i < delayms; i++) {     // wait for 'delayms' millisecs
    digitalWrite ( AUDIO_OUT, HIGH);
    if (Loud)  digitalWrite ( AUDIO_NEG, LOW);
    delayMicroseconds(840);    // tried: 700 600 400 456 best? 700 825 840
    digitalWrite ( AUDIO_OUT, LOW);
    if (Loud)  digitalWrite ( AUDIO_NEG, HIGH);
    delayMicroseconds(560);   // tried: 550 400 266 210 best? 500 550 560
    refreshDisp();  // just spent ~1 ms, lets refreshDisp now
  }
  digitalWrite ( AUDIO_OUT, LOW);;
  if (Loud)  digitalWrite ( AUDIO_NEG, LOW);
}

// -----------------------
void dahAudio(unsigned short int delayms) { // lower freq ( w/ longer ON duty cycle for Piezo)
  delayms = (2 * delayms) / 3;
  for (int i = 0; i < delayms; i++) {     // wait for delayms ms
    digitalWrite ( AUDIO_OUT, HIGH);
    if (Loud)  digitalWrite ( AUDIO_NEG, LOW);
    delayMicroseconds(870);    // 900
    digitalWrite ( AUDIO_OUT, LOW);
    if (Loud)  digitalWrite ( AUDIO_NEG, HIGH);
    delayMicroseconds(580);    // 600
    refreshDisp();  // just spent ~1 ms, lets refreshDisp now
  }
  digitalWrite ( AUDIO_OUT, LOW);;
  if (Loud) digitalWrite ( AUDIO_NEG, LOW);
}


// -----------------------
void playMorseCode(byte ditdahs, int pace) {
  // all MSBits that are the same are discarded, whats left is 1=dit 0=dah, MSbit is first in time, LSB last
  short int ditPeriod;
  byte i = 1;
  byte bitState;

  ditPeriod = pAveraging / 8;  // set according to their current pace
  ditPeriod = 8 * ditPeriod / 10; // but 80% timing, for 1.25 the speed
  if (pace) ditPeriod = pace;  // unless passed in

  if (ditdahs == 0) return;
  while ((ditdahs << i & 0x80) == (ditdahs & 0x80)) i++;
  for (int b = i; b < 8; b++) {
    if (ditdahs << b & 0x80) { // this bit is a 'dit'
      brightOne = 7;
      ditAudio(ditPeriod);
      brightOne = 0;
      refreshWait(ditPeriod);
    } else {  // its a dah
      dim[7] = dim[8] = true;
      dahAudio(3 * ditPeriod);
      dim[7] = dim[8] = false;
      refreshWait(ditPeriod);
    }
    scanBtns();  // needed to keep button state upto date (for btn4 esc)
  }
}


// -----------------------
void playStr(char *wrd, int pace) {          // ment for message announcements
  float ff = farns;
  char c;
  byte ditdahs;
  int ditPace;
  byte dimF;

  dimF = dimFlash;
  dimFlash = 0; // stop position flash in MC_Practice during this word output
  refreshWait(10);

  ditPace = ditPeriod;
  if (pace) ditPace = pace;  // unless passed in

  for (int i = 0; i < strlen(wrd); i++) {
    ditdahs = Ascii2dida(wrd[i]);
    if (wrd[i] == '_') ditdahs = 0;
    //Serial.print(wrd[i]); Serial.print(ditdahs);
    if (ditdahs == 0) delay(2 * ditPace * ff); // handle timing for the likes of a space
    playMorseCode(ditdahs, ditPace);
    if (i+1<strlen(wrd)) {  // more coming
      //    refreshWait(3*ditPace*ff-ditPace);   // +2p for between character
      refreshWait(2*ditPace*ff);   // +2p for between character
    }
    scanBtns();
    if (btn4) return;
  }
  //  refreshWait(3*ditPace*ff);  // add 3+p more time to make up 7p between word period
  //  now caller's responsibility
  dimFlash = dimF;
}

// ---------- 
void reportRate() {
  float wpm;
  wpm = ((1000.0 * 60.0) / (240.0 * ditPeriod / 5.0));
  //Serial.print(" avg. Period: "); Serial.println(pAveraging/4);
  Serial.println();
  Serial.print(" Keying rate (in WPM): "); Serial.print(wpm);
  if (farns == 1) Serial.println();
  else Serial.println("  + Farnsworth spacing");
  showLevel(wpm + 0.5);
  refreshWait(2000);
  clearDisp();
}

//==================  funtional building blocks to support Morce Code I/O
// ---------- Morse code support functions ------------
byte fetchMorseChar(bool echo) {
  debounceDelay = 5; // do extra button debouncing
  if (oneKey) return (fetchMorse_1Key(echo)); // uses One button keying, based on duration of presses
  //  else        return(fetchMorse_2Key(echo));  // uses basic Two button keying
  else        return (fetchMorse_22Key(echo)); // uses basic Two button keying improved p timing
  //  else        return(fetchMorse_2A_Key(echo));  // uses Two button with fixed timing & Auto keying
}

byte fetchMorse_1Key(bool echo) { // get one char (ditdah set) using single button keying (straight CW key)
  byte mChar=0;
  int  p=0, ditPs;
  unsigned long pt;
  byte dit;
  //static bool diTurn = true;
  //int lastavgDah;

  // update timing used prior to each whole character
pAveraging=(avgDit+(avgDah/3))/2;
//pAveraging=(avgDit+avgDah)/4;
  avgPeriod = pAveraging/4; // the running average maintains 4*resolution desired
  ditPeriod=avgPeriod/2;    // avgPeriod=2*ditPeriod  aka  the average between dit and dah period   

  while(!(btn1 || btn3 || btn4)) {scanBtns(); refreshWait(1);} // wait for button press
  while (btn1) {
    p=0;
    pt = millis();  // time the period
    while (btn1) {
      dit_dah(2);   // output audio while key pressed (done for 2, sounds better than for 1)
      scanBtns();
    }
    p = millis()-pt;
    if (p>(5*ditPeriod)) p=5*ditPeriod;  // curb extream dah  (sb: 4 or 5)

    clearDisp();
 //    dit = (p<avgPeriod);
 //    dit = (p<(9*avgPeriod/10));  // try 1.8 ditPeriod the break point between a dit & a dah
    dit = (p<(7*avgPeriod/8)); // 1.75 of a dit
    // 1.8 is +80% of a dit, and -40% of a dah
    // 1.67 if +67% of a dit, and -45% of a dah
    ditPs=(p+ditPeriod/2)/ditPeriod;
    //lastavgDah=avgDah;

    if (p>(avgPeriod/8)) { // filter out spurious spikes
      if (mChar==0) { // setup first bit
        mChar = (dit)? 0x01 : 0xFE;
      } else {    // shift in other bits
        mChar = (mChar<<1) | dit;
      }
      // add into the running average 'pAveraging' (but not ones from left field)
      if (dit && (2*p>ditPeriod)) { // removed diTurn &&
        if (p>(ditPeriod+ditPeriod/2)) p=(ditPeriod+ditPeriod/2); // temper increases to avgDit
        pAveraging = (7*pAveraging/8) + p;
        avgDit = (7*avgDit/8) + p;
        //diTurn = !diTurn; // average in = # of dits & dahs
      }
      ///    if (!dit && p<(5*ditPeriod)) { // removed !diTurn &&
      if (!dit) { // removed !diTurn &&
        if (p>(4*ditPeriod)) p=(4*ditPeriod);
        if (p<(2*ditPeriod)) p=(2*ditPeriod); // temper decrease to avgDah
        pAveraging = (7*pAveraging/8) + (p/3);
        avgDah = (7*avgDah/8) + p;
        //diTurn = !diTurn;
      }
      //if (ditPs>4) avgDah=lastavgDah;  // dah too extream
    }
    //else  Serial.print("'"); // mark the rejected spurious spike

    p=0;
    while (Btn==0) {  // wait for additional dit-dahs (or other)
      scanBtns();
      delay(1);
// the following needs to balance between an intervle long for inter dit-dah (1p) and short for inter char (3p)
//      if (++p>(7*ditPeriod/3)) {  // require only 2.3 (3 expected be the average)
//      if (++p>(2*ditPeriod)) {
      if (++p>(11*ditPeriod/5)) {  // require only 2.2 (3 expected be the average),  make "."  .-.-.- easier
        break;          // done with this char
      }
    }
  }

  if (btn3 || btn4) return 0;
  if (echo) {
    set_didaStr(mChar, didaStr);
    Serial.print("  ");
    Serial.print(didaStr); Serial.print(" ");
  }

  // TBD ??
  // if (func==1 && ditPs>4 && dida2Ascii(mChar)=='T') {
  //   mChar=0; 
  //   Serial.print(char(0x08));Serial.print(' ');Serial.print(char(0x08));
  // } else 
    if (prtChr) Serial.print(dida2Ascii(mChar));  // always output the alpha character (patched)
 
  //TBD debugDit(mChar);
  return (mChar);
}


// -----------------------
byte fetchMorse_2Key(bool echo) { // using Two button keying (or a dual lever paddle)
  return (fetchMorse_22Key(echo));
}
// -----------------------
byte fetchMorse_22Key(bool echo) { // using Two button keying (or a dual lever paddle)
  // dit & dah differ in sound, but duration is simply based on button press duration.
  // TODO: Not throurghly valided, may need/want adjustments to period timing
  //  static short int ditPeriod;
  byte mChar = 0;
  int  p = 0, p2 = 0, pCnt = 0;
  bool dit;
  byte msn;
  bool toggle = false;
  byte ditdaCnt = 0;

  avgPeriod = pAveraging / 4; // the running average maintains 4*resolution desired
  ditPeriod = avgPeriod / 2;  // avgPeriod=2*ditPeriod  aka  the average between dit and dah period

  while (!Btn) {
    scanBtns();  // wait for button press
    refreshWait(1);
  }
  while (btn1 || btn2) {
    ditdaCnt++;
    p += p2;
    dit = btn1;
    t0 = millis();
    while (Btn) {
      scanBtns();
      if (btn1) {   // produce sound while pressing
        brightOne = 7;
        ditAudio(1);
      } else {
        brightOne = 7; litLED = 8;
        dahAudio(2);
      }
      toggle = !toggle;
      p++;
      clearDisp();
    }

    if (mChar == 0) { // setup first bit
      mChar = (dit) ? 0x01 : 0xFE;
    } else {    // shift in other bits
      mChar = (mChar << 1) | dit;
    }
    // count it into the running average 'pAveraging'
    if (dit) pCnt += 2;
    else pCnt += 3; // compromise, as it takes 2 to key but 4 to reproduce

    p2 = 0;
    while (Btn == 0) { // wait for additional dit-dahs
      scanBtns();
      delay(1);
      if (++p2 > (2 * ditPeriod)) {
        break;      // done here, end of a char
      }
    }
    if (dit) {  // just had a dit + an inter period
      pAveraging = (3 * pAveraging / 4) + (millis() - t0); // provide tracking of avg. for 2p
      avgPeriod = pAveraging / 4;
      ditPeriod = avgPeriod / 2;
    }
  }

  if (btn3 || btn4) return 0;
  if (echo) {
    set_didaStr(mChar, didaStr);
    Serial.print("  ");
    Serial.print(didaStr); Serial.print(" ");
  }
  if (prtChr) Serial.print(dida2Ascii(mChar));  // always output the alpha character

  if (DEBUG) {
    Serial.print("   Morse Char: ");  Serial.print(mChar);
    Serial.print("  0x"); Serial.print(mChar, HEX);
    Serial.print("    ave Period: ");  Serial.println(2 * ditPeriod);
    Serial.print(" Traking Period: ");  Serial.println(pAveraging / 4);
  }

  return (mChar);
}

// -----------------------
char dida2Ascii(byte ditdahs) {
  byte ncode, ascii, msb, ndidas, i;
  const char nc2A[]  = "~EAIWRUSJP\nL~FVH~~QZYCXBOGKDMNT~";
  //  const char dida5[] = "09 8   7 (   /=61    + &2   3 45";
  //  const char dida6[] = " '   @ :  . ,\"    )?! $       - ";
  const char dida5[] = "09~8~~~7~(~~~/=61~~~~+~&2~~~3~45";
  const char dida6[] = "~'~~~@~:~~.~,\"~~~~)?!~$~~~~~~~-~";
  static bool lastP = false; // last char processed was a () type

  //Serial.print(" ditdah code: ");  Serial.println(ditdahs);
  if (ditdahs == 0) return 0;

  // strip off all most significant extra 1s bits
  i = 1;
  msb = (ditdahs & 0x80);
  while ((ditdahs << i & 0x80) == msb) {
    i++;
  }
  ndidas = 8 - i;
  ncode  = ditdahs & 0x1F;
  // decode according to its length
  if (ndidas < 5)       ascii = nc2A[ncode];
  else if (ndidas == 5) ascii = dida5[ncode];
  else if (ndidas > 5)  ascii = dida6[ncode];

  if (!lastP) {
    if (ascii == ')' && pOpen) ascii = '(';
    if (ascii == '(') pOpen = false;
    //was    if (ascii==')' || ascii=="\n") pOpen=true;
    if (ascii == ')' || ascii == 0x0A) pOpen = true;
  }
  if (ascii == '(' || ascii == ')') lastP = true;
  else lastP = false;
  return (ascii);
}

// -----------------------
byte Ascii2dida(char Achar) {
  byte ncode, i;
  const char nc2A[] = "~EAIWRUSJP~L~FVH~~QZYCXBOGKDMNT~~";
  int alen;

  Achar = toupper(Achar);
  //  move i-index past padding bits
  for (i = 1; i < 32; i++) {
    if (nc2A[i] == Achar) break;
  }
  if (i < 32) {
    ncode = (i < 0b10000) ? i : i + 0b11100000;
  } else {  // handle #s, punc. & spec. chars
    if (Achar == '[' || Achar == '{')  Achar = '(';
    if (Achar == ']' || Achar == '}')  Achar = ')';
    alen = sizeof(nonAlpha);  // sizeof(char)=1
    for (i = 0; i < alen; i++) {
      if (nonAlpha[i] == Achar) break;
    }
    ncode = nonAlpha_dida[i];
  }
  return (ncode);
}

// --------------------------------------------
void set_didaStr(byte ditdahs, char ddStr[]) { // returns string representation in global var: 'didaStr'
  // all MSBits that are the same are discarded, whats left is 1=dit 0=dah, MSbit is first in time, LSB last
  byte i = 1;
  byte bitState;

  if (ditdahs == 0) return;

  while ((ditdahs << i & 0x80) == (ditdahs & 0x80)) i++;
  int c = 0;
  for (int b = i; b < 8; b++) {
    if (ditdahs << b & 0x80) { // this bit is a 'dit'
      ddStr[c++] = '.';
    } else {  // its a dah
      ddStr[c++] = '-';
    }
  }
  ddStr[c++] = 0;
}

int pickOne(int min, int max) {   // returns value in range min - max-1
     // each value occuring 2 times of 2* the range, with 1/2 the repeats given by random(). Good for pseudo random practicing exercises
     // especially when invovling a limited # of items, like alphabets, character sets, training routines, exam questions etc.
  static unsigned int si=0;
  return (min + PRIG(si++, gameID, 2*(max-min))/2); 
}

// ***************** used by Mega_Games.ino to make up:  Group 6  ************************
// ----------- Morse Code Practice & Exercise samples ------------------

void MC_freeForm() {  //  -------- allow free form Morse code practice until btn4 pressed
  // mostly for Beginers, and 2 Key M.C. operation is not supported
  byte ditdahs;
  char chrs = 0;
  int p, ms;
  float wpm;

  prtChr=false;
  while (!btn4) {
    if (btn1) {
      ditdahs = fetchMorseChar(0);  //echo);   // get dits & dahs appearing to form the MC of a character
      if (ditdahs != 0) {
        Serial.print(dida2Ascii(ditdahs));  // output the alpha character
        chrs++;
        p = 0;
      }
      litLED = (chrs - 1) % 8 + 1;
    }
    if (btn3) {  // tap B3: faster,  short hold B3: reduce speed
      ms=0;
      while (btn3) {
        scanBtns();
        refreshWait(1);
        ms++;
      }
      if (ms<200) ditPeriod = (6 * ditPeriod + 3) / 7; // pickup the pace 16+% faster
      else        ditPeriod = (7 * ditPeriod + 2) / 6; // slow down the pace 16+% slower
      wpm = ((1000.0 * 60.0) / (240.0 * ditPeriod / 5.0));
      Serial.print(" Keying rate: "); Serial.println(wpm);
      playStr((char*)"R", 0);  // .-.  2p for each
    }

    // ---------- checking for end of word
    if (p < (5 * ditPeriod)) {
      if (++p == (5 * ditPeriod)) Serial.print(" "); // give 3+5 more than 7p
    }
    scanBtns();
    refreshWait(1);
  }
  pAveraging = 8 * ditPeriod; // update their standard rate
  reportRate();
}


char AZ_Words[]  =  ",Apple,Baby,Cat,Dog,Egg,Fish,Goat,Horse,Ice,Jug,Kite,Lion,Monkey,Nose,Owl,Pig,Queen,Rabbit,Snake,Tiger,Ugly,Vase,Whale,Xmas,Yellow,Zebra,";
// !!! use 3 char words, short version to keep free dynamic memory >300 !!!
//char AZ_Words[] = ",Ape,Box,Cat,Dog,Egg,Fox,Gum,Hat,Ice,Jug,Key,Log,Man,Nut,Owl,Pig,Que,Rag,Sky,Tux,Urn,Van,Web,Xor,Yap,Zip,";
//char AZ_Words[] = ",Ape,Bat,Cat,Dog,Egg,";    // for quick testing
char NATO_Words[] = ",Alpha,Bravo,Charlie,Delta,Echo,Foxtrot,Golf,Hotel,India,Juliet,Kilo,Lima,Mike,November,Oscar,Papa,Quebec,Romeo,Sierra,Tango,Uniform,Victor,Whiskey,X-ray,Yankee,Zulu,";
// -----------------------   A-Z Morse Code Please
void A2Z_Morse_Exercise() { // Exercise keying in A-Z in order. or alphabet phonetic words Apple-Zebra or NATO Phonetics
  //  char alpha[] = "ZABCDEFGHIJKLMNOPQRSTUVWXYZA";
  byte ditdahs;  // all MSBits that are the same are discarded, whats left is 1=dit 0=dah, MSbit is first in time, LSB last
  byte dida;  // expected
  char cinput, c;
  byte correct = 0;
  static bool echo = true;
  byte opt,ccnt, p=1;
  byte err=0;

  opt = get_Mode(3); // opt 1: letters 'A'-'Z'     2: alphabet Phonetics     3: NATO Phonetics
  ccnt = (opt==1)? 26 : (opt==2)? strlen(AZ_Words)-1 : strlen(NATO_Words)-1;

  pAdj = false; // maintain Keying rate
  // display & btn press is cleared prior
  delay(100);
  for (int i = 1; i <= ccnt; i++) {
    if (opt==1)       c=alpha[i];    // ---- get next character 
    else if (opt==2)  c=AZ_Words[i];
    else              c=NATO_Words[i];
    Serial.println(); Serial.print(c);
    if (c==',') {
      if (echo) {delay(75); beep(50,50); beep(175);}
      p++;
      if (err==0) correct++;
      err=0;
      continue;
    }

    dimFlash = (p - 1) % nleds + 1;
    while (1) {
      refreshWait(1);
      if (btnPressed() && (btn1 || btn2)) {    // Button has been pushed
        ditdahs = fetchMorseChar(true);     // get one char response
        //refreshWait(200);  // ?

        cinput = dida2Ascii(ditdahs);
        dida = Ascii2dida(c);
        if (cinput == toupper(c)) { // he got it right
          if (opt==1) correct++;
          if (echo && opt==1) playMorseCode(ditdahs, 0);
        } else {
          err++;
          while (ditdahs != dida && ditdahs != 0) { // give them the correct ans, & do so until they get it right
            brt[7] = brt[8] = true;
            buZZ(500, 250);
            brt[7] = brt[8] = false;  refreshWait(150);
            playMorseCode(dida, 0);
            dimFlash = (p - 1) % nleds + 1; // reestablish
            ditdahs = fetchMorseChar(true);
            scanBtns();
            if (btn4) return;
          }
          if (echo && opt==1) playMorseCode(ditdahs, 0);
        }
        clearDisp();
        //            delay(500);  // ? whats advisable here
        break;
      }
      if (btn3Pressed) echo = !echo;
      if (ESC || btn4) return;
    }
    if (opt==1) p++;
  }

  Serial.print("\ncorrect: "); Serial.print(correct);
  Serial.print("  out of  "); Serial.println(26);
  showLevel(correct);
  refreshWait(3000); clearDisp();
}


// -----------------------
void Morse_AB_Prompt() { // Follow random letter with the MC for the letter following it alphabetically
  //  const char alpha[] = "ZABCDEFGHIJKLMNOPQRSTUVWXYZA";
  int si=0, ci;
  int correct;
  char cinput;
  byte dida, ditdahs;
  byte ntests = 13; // all 26 letters get exercised in two sessions
  byte glowLED;
  byte opt;
  byte err;
  char srch[]=",?";
  int wi;
  char pC, tC;

  pAdj = false; // maintain Keying rate
  opt = get_Mode(3); // opt 1: AB next letter test 'A'-'Z'     2: alphabet Phonetics     3: NATO Phonetics

  glowEnable = true;
  pAdj = false; // maintain Keying rate
  while (1) {
    refreshWait(1000);
    correct = 0;
    for (int n = 1; n <= ntests; n++) {
        err=0;
        glowLED = (n - 1) % nleds + 1;
        glow[glowLED] = true; // mark test progress
        dimIt = glowLED;  // dimIt softer
 
        while (Btn) scanBtns(); // wait out old btn presses
        ci = PRIG(si++,gameID,26)+1; // use Pseudo Random Index Generater & let si keep on going
        pC=alpha[ci];                 // set Prompt Char
        tC=(opt==1)? alpha[ci+1] : pC; // set Target Char
        if (opt>1) {
          srch[1] = pC;
          if (opt==2) wi = (strstr(AZ_Words, srch)) - AZ_Words ;
          if (opt==3) wi = (strstr(NATO_Words, srch)) - NATO_Words ;
        }

        dida = Ascii2dida(pC);
        set_didaStr(dida, didaStr);
        playMorseCode(dida, 0);              // ---- sound one out
        Serial.println();
        Serial.print(didaStr);  Serial.print(" ");  Serial.print(pC);

     do {
       if (opt==2) {pC = tC = toupper(AZ_Words[++wi]);}
       if (opt==3) {pC = tC = toupper(NATO_Words[++wi]);}
       if (tC==',') break;

        ditdahs = fetchMorseChar(true);  // ----  get their response
        if (btn4) return;
        if (btn3) {
          si--;  n--;       // replay the prompt char
          refreshWait(750);
          break;   // go back and do it again
        }
        cinput = dida2Ascii(ditdahs);
        if (cinput == tC) {  // he got it right
          //correct++;
        } else {
          err++;
          while (cinput != tC) { // Prompt them, until they get it right
            if (ditdahs != 0) {
              brt[7] = brt[8] = true;
              buZZ(500, 500);
              brt[7] = brt[8] = false; refreshDisp();
            }
            if (btn3) delay(750);
            while (Btn) scanBtns();  // clear out btns
            dida = Ascii2dida(pC);
            playMorseCode(dida, 0);
            ditdahs = fetchMorseChar(true);
            if (btn4) return;
            cinput = dida2Ascii(ditdahs);
          }
        }
        clearDisp();
        if (ESC || btn4) return;
      } while (opt>1);
      if (!err) correct++;
      while(err) {Serial.print("*"); err--;}
      refreshWait(750);
    }

    Serial.print("\ncorrect: "); Serial.print(correct);
    Serial.print("  out of  "); Serial.println(ntests);
    showLevel(correct);
    refreshWait(3000); clearDisp();
  }
}

// -----------------------
void MC_Relay_exercise() { // Write down the letters of a 'word' as you hear them. Then try to key them in (as if to relay)
  const char SpChar[] = "01234567890.,?/'-.,?/'-:\"+!=@$)&\n";
  char dd[10];  // stores ditdah given out
  byte ci, li, wlen, numW, mini, maxi;
  int i, p, dl;
  static int opt = 1;
  char dida;
  char c, cinput;
  byte correct;
  static int ditPace;
  static bool adjd = false;
  float wpm;

  if (!adjd) ditPace = 5 * ditPeriod / 6; // use 5/6 his current dit period, play at 1.2x keying speed
  //// TBD pAdj = false; // maintain Keying rate
  // note: once manually overriden, new speed will get used
  wpm = ((1000.0 * 60.0) / (240.0 * ditPace / 5.0));
  Serial.print(" Audio out rate to be used: "); Serial.println(wpm);

  opt = get_Option(5, opt);
  if (ESC || opt == 0) return;
  mini = 1; // default
  //  determin the char set to be used
  if      (opt == 1) {
    mini = 1;  // Alpha
    maxi = 26;
  }
  else if (opt == 2) {
    mini = 1;  // Alpha + Nums + punctation . thru +
    maxi = 51;
  }
  else if (opt == 3) {
    mini = 27;  // Nums + punctation . thru +
    maxi = 51;
  }
  else if (opt == 4) {
    mini = 37;  // just punctation-special chars
    maxi = 58;
  }

  wlen = (opt == 5) ? 5 : 3; // was 5:4
  numW = (opt == 5) ? 4 : 4; // was 4:3

  clearDisp();
  while (!btn4) {
    correct = 0;
    //    clearDisp();
    while (btnDown()) refreshWait(1);

    /***
    // ------  wait for two second, mean while btn2&3 can be used to adjust speed
    dl = 2000; // delay for 2secs
    while (dl--) {
      scanBtns(); refreshWait(1);
      if (btn4) return;
      if (btn2Pressed || btn3Pressed) {
        if (btn3) ditPace = (5 * ditPace + 3) / 6; // pickup the pace 20% faster
        else      ditPace = (6 * ditPace + 2) / 5; // slow down the pace 20% slower
        adjd = true;
        wpm = ((1000.0 * 60.0) / (240.0 * ditPace / 5.0));
        Serial.print(" Audio out rate: "); Serial.println(wpm);
      }
      if (BtnPressed) {
        dl = 3000; // extend waiting
        playMorseCode(Ascii2dida('P'), ditPace); // give a 'P' for reference
      }
    }
    ***/  delay(1200);

    for (int w = 1; w <= numW; w++) {
      clearDisp();
      if (opt == 5) maxi = (random(3) == 0) ? 26 : 42;
      for (li = 0; li < wlen; li++) { // generate random letters of a psuedo word
        //ci = random(mini, maxi + 1); // get the character form the set he selected
        ci = pickOne(mini, maxi + 1);  // now, uses the PRIG func for better behavior for test matterial generation
        if (opt == 5 && li == 4) ci = random(37, 58); // add an extra spc. char
        if (ci < 27) {
          c = alpha[ci];
          Serial.print(c);
          dida = Ascii2dida(c);
        } else {
          i = ci - 26;
          c = SpChar[i];
          Serial.print(c);
          if (i > 16) i -= 6; // adjust index to skip over repeated chars in SpChar[]
          dida = nonAlpha_dida[i];
        }
        dd[li] = dida;

        playMorseCode(dida, ditPace);
        if (btn4) return;
        if (ditPace > 120)  delay(2 * ditPace); // std timing between char pause
        else delay(2 * ditPace * farns); // give students +time to recallet with farnsworth
        if (farns == 1) delay(3 * ditPace); // still give +time to recallet std speed chars, they just heard

        set_didaStr(dida, didaStr);
        Serial.print(" "); Serial.print(didaStr); Serial.print("    ");
      }

      Serial.println();
      dimFlash = 1;
      while (!(btn1 || btn2)) {  // wait for start of input
        refreshWait(1);
        scanBtns();
        if (btn4) return;
        if (btn3Pressed) {
          refreshWait(1000);
          for (li = 0; li < wlen; li++) { // replay the 'word'
            playMorseCode(dd[li], ditPace);
            refreshWait(2 * ditPace * farns);
          }
        }
      }
      for (li = 0; li < wlen; li++) { // get their reponse to the psuedo word given
        dimFlash = li + 1;
        dida = fetchMorseChar(true);
        if (ESC || btn4) return;
        Serial.print("  ");
        if (dida == dd[li]) {
          correct++;
          dim[li + 1] = true;
        }
        else buZZ(150);
      }
      dimFlash = 0;
      refreshWait(1000);
      Serial.println();
    }
    Serial.print("\ncorrect: "); Serial.print(correct);
    Serial.print("  out of ");  Serial.println(wlen * numW);
    if (correct == 0) correct = 1; // credit for participation
    showLevel(correct);
    refreshWait(3000); //clearDisp();
  }
}

// -------------
void MC_Receive_Practice() { // Practice listening to and making out MC characters at your own pace.
// came from 'MC_Practice' ...  removed old mode3&4 then added Two random letters continuous mode. old #2 became #3
  byte opt, mode;
  char wrd[30], tmpw[30];
  char *cptr;
  int strInx;
  byte ci, wlen, errs, lErrs, ccnt;
  int si=0; // working word-ish selection index
  bool errW;
  int i, ri, w, p;
  byte ditdahs;
  char cinput, c;
  float wpm, wpmTkr;
  unsigned long t, t0;
  const int hiSpd = 19; // to use simplier words when working at lower speeds
  bool test = true; // normally its a prompted test, else its free form sending (CW keying)
  bool fini;
  int ditPace;
  byte offset;

  // old Modes:  1) RCV word at a time   2) RCV a series of words
  // Modes:  1) RCV word at a time   2) two 2 random leters at a time non stop   3) RCV a series of words
  // words are all randomly selected
  Serial.print(" Mode #");
  mode = get_Mode(3); Serial.println(mode);
  if (mode == 0) return;

  //numWrds=strlen(wordStr)/5;  // for approximate
  numWrds = 500;
  offset = (gameID % 3) *4; // offset for stepping thru selections

  while (1) { // provide back to back operation, for long practice sessions
    ditPace = 5 * ditPeriod / 6; // set audio out higher than their keying rate
    if (ditPace > 102) ditPace = 102; // maximum of 102 for minimum. 102 for 12 WPM (92 for 13+, 77 for 16 WPM)
    wpm = ((1000.0 * 60.0) / (240.0 * ditPace / 5.0));
    Serial.print("\nStarting char output WPM rate: "); Serial.println(wpm);

    fini = false;
    ci = 0;
    ccnt = 0;
    errs = lErrs = 0;
    w = 0;
    wpmTkr = 0;
    clearDisp();
    refreshWait(1000);

    while (ccnt < 60) { // practice test is at least 60 chars
      pOpen = true;
      dimFlash = 1 + (w & 0x07);
      refreshDisp();

      if (mode==2) {
        wrd[0] = 'A' + PRIG(si++,gameID, 26);  // select 2 ramdon letters
        wrd[1] = 'A' + PRIG(si++,gameID, 26);  // use PRIG instead of random() for no dups and all letters exercised within 26 used
        wrd[2] = 0;
        wlen=2;
      } else {
        do {  //    randomly select a 'word'
          //strInx = random(0, strlen(wordStr) - 26); // randomly select an area in which to look for a 'word' [generated ~5% repeats]
          strInx = PRIG(si++, gameID, (strlen(wordStr)-8)/12) *12 + offset;  // now, use PRIG to get random selection minimizing repeats
                   // goes though ~300 unique selection  for each of 3 offsets (0,4,8) there is ~1/3 overlap between the 3 groups
          strncpy_P(wrd, (char *) & (wordStr[strInx]), 26 );
          wrd[26] = 0;  // 1 beyond 0-25 put a string terminator

          strcpy(tmpw, strchr(wrd, ',') + 1); // pull out a word
          cptr = strchr(tmpw, ',');
          if (cptr == NULL) continue;
          wlen = (int) (cptr - tmpw);
          strncpy(wrd, tmpw, wlen);
          wrd[wlen] = 0;
          //        Serial.print("  wlen is "); Serial.print(wlen);  Serial.print(" : "); Serial.println(wrd);
          //        Serial.print(" strInx is "); Serial.print(strInx);  Serial.print(" : "); Serial.println(wrd);
        } while ( wlen < 1 || (wpm < hiSpd && wlen > 6) || (wpm < (4+hiSpd) && wlen > 8));
        // in the 900 element array there are 635 entries with 3-6 characters, 812 with 8 or less characters, 88 words more than 8
      }
      ccnt += wlen;

      wpm = ((1000.0 * 60.0) / (240.0 * ditPeriod / 5.0));
      if (wpmTkr == 0) wpmTkr = 10 * wpm;
      wpmTkr = (9 * wpmTkr / 10 + wpm); // track & average over a practice round

      if (test) {
        Serial.print(" ");
        Serial.print(wrd);
      }

      for (i = 0; i < strlen(wrd); i++) {
        c = wrd[i];   // use the same ditdah code for all brackets & braces
        if ( c == '[' || c == '{' || c == '(' || c == '}' || c == ']' ) wrd[i] = ')';
      }

      if (w == 0) t0 = millis();
      playStr(wrd, ditPace);
      if (mode > 1) refreshWait(3 * ditPace * farns); // add 3+p more time to make up 7p between word period
      if (mode==2)  refreshWait(3 * ditPace * farns); // double time between "words"

      btnsNow();  // scan for new button state
      if (btn4) fini = true;
      else if (mode == 3 && Btn) {           // go into Pause state
        set_LEDs(0x30, 3);
        while (Btn) {
          refreshWait(25);
          scanBtns();
        }
        set_LEDs(0x30, 0);
        refreshWait(1500);   // give time to return to paper & pencil
      }

      if (mode == 1) { // handle one word at a time
        while (!(btn1)) { // ---------- wait to CW keying to begin
          refreshWait(1);
          scanBtns();
          if (btn4) {
            fini = true;
            break;
          }
          if (btn3) {
            POV_display(wrd);   // do a POV display of the MC word
          }
          while (btn3) scanBtns();  // cause btn3 to toggle POV on/off
          if ((btn2 && mode == 1)) {
            delay(1000);
            playStr(wrd, ditPace);
          }
        }
      } else {
        scanBtns();
        if (btn4 && mode == 2) return;
      }

      if (fini) break;
      dimFlash = 0;
      litLED = 0; refreshDisp();
      if (mode == 1) refreshWait(750); // moment to get ready to write it down
      w++;
    }

    t = (millis() - t0);
    //Serial.print("\n Char cnt: "); Serial.print(ccnt); Serial.print("   time(msecs): "); Serial.print(t); Serial.print("   mins: "); Serial.print(t/60000.0);
    if (mode == 1) t = t - 8; // -(0.75*11) for the waits the code injected
    wpm = ((ccnt - 1 + w * farns) / 5.0) / (t / 1000.0 / 60.0); // +12*farns is adj for alowing extra time between letters & words
    Serial.println(); Serial.print("   over all WPM: "); Serial.println(wpm);
    showLevel(wpm);
    refreshWait(2000);
    while (!btnPressed()) {
      delay(1); // wait for button press
    }
    if (btn4) return;  // B4->exit, else go again
    while (btnDown()) ;        // then wait for release
  }
}

// -----------------------
void Tx_Line_test() {  // Practice Transmitting, Key in complete lines as seen on the Monitor screen.
  // btn 3 toggles Easy mode, where bad spacing is tollerated
  // each time you complete a line with no errors 10 characters is added to the test lines (upto 80 chars)
  char wrd[30], tmpw[30];
  char line[100];
  static const char ending[] = ".,?!-/+.,?";
  byte ci, wc, wlen, wcnt, errs, ccnt, llen;
  int wi;
  byte i, xt;
  int p, pspace;
  byte ditdahs;
  char cinput;
  float wpm, wpmTkr;
  int wpm_min=0, wpm_max=0;
  byte chrs;
  bool lerr, ptime;
  static bool easy = true; // true= don't penalize poor spacing
  static byte maxC = 10;
  unsigned long t, pauseT;
  char testC;
  char *cptr;
  int strInx;
  //  const byte hiSpd=18;  // to use simplier words when working at lower speeds
  const byte hiSpd = 14;
  byte offset;
  int si=0; // working word-ish selection index

  //algoChkSum();   // used to verify that the Miller Shuffle Algo 'PRIG' was ported correctly (sb: 6447339)

  randomSeed(randSeed = micros() / 4); // micros() resolution is ~ 4us
  Serial.print("random Seed:  "); Serial.print( randSeed);

  prtChr = false; // prevents double printing MC chars

  wpm = ((1000.0 * 60.0) / (240.0 * ditPeriod / 5.0));
  if (wpm > 17) easy = false;
  //  Serial.print("Starting char WPM rate: "); Serial.println(wpm);
  wpmTkr = 6 * wpm;
  delay(200);  // avoid button bounce on entry
  offset = (gameID % 3) *4; // offset for stepping thru selections

  while (!btn4) {
    i = 0; ci = 0;
    wpmTkr=0;
    ccnt = 0;
    wcnt = 0;
    chrs = 0;
    pauseT = 0;

    while (ccnt < maxC) {  // predetermine 'word's for line totaling <80 chars

      do {  //    randomly select a 'word'
        //strInx = random(0, strlen(wordStr) - 26); // randomly select an area in which to look for a 'word' [generated ~5% repeats]
        strInx = PRIG(si++, gameID, (strlen(wordStr)-8)/12) *12 + offset;  // now, use PRIG to get random selection minimizing repeats
                   // goes though ~300 unique selection  for each of 3 offsets (0,4,8) there is ~1/3 overlap between the 3 groups
        strncpy_P(wrd, (char *) & (wordStr[strInx]), 26 );
        wrd[26] = 0;

        strcpy(tmpw, strchr(wrd, ',') + 1); // pull out a word
        cptr = strchr(tmpw, ',');
        if (cptr == NULL) continue;
        wlen = (int) (cptr - tmpw);
        strncpy(wrd, tmpw, wlen);
        wrd[wlen] = 0;
        //        Serial.print("  wlen is "); Serial.print(wlen);  Serial.print(" : "); Serial.println(wrd);
        //        Serial.print(" strInx is "); Serial.print(strInx);  Serial.print(" : "); Serial.println(wrd);
      } while ( wlen < 1 || (wpm < hiSpd && wlen > 6) || (wpm < 4 + hiSpd && wlen > 8));
      // in the 900 element array there are 635 entries with 3-6 characters, 812 with 8 or less characters, 88 words more than 8

      for (i = 0; i < wlen; i++) {
        line[ci + i] = wrd[i]; // crude string copy
      }
      ci = ci + wlen;
      line[ci++] = ' ';
      ccnt += wlen;
      wcnt++;
    }

    line[ci - 1] = ending[random(0, strlen(ending))];
    line[ci] = 0x00; // mark end of string
    llen = strlen(line);
    Serial.print("\n "); Serial.println(line); Serial.print(" ");

    ci = 0;
    wi = 0;
    ccnt = 0;
    t0 = 0;
    errs = 0;
    xt = 0;
    lerr = false;

    ptime = true; // allow time to pass before keying begins
    while (ci < llen) {
      cinput = '#';
      p = 0;
      while (cinput == '#') { // ---------- get next MC character from user
        dimFlash = 1 + (wi & 0x07);
        refreshWait(1);
        if (btnPressed()) {
          if (btn1 || btn2) {   // ---- they're keying in MC
            if (t0 == 0) t0 = millis();
            ditdahs = fetchMorseChar(false);   // get dits & dahs appearing to form the MC of a character
            t1 = millis();
            pauseT = millis();
            cinput = dida2Ascii(ditdahs);\
            if (easy && line[ci] == ' ' && (p > pspace/3)) {
              ci++;
              Serial.print(' ');
            }
          }
          if (btn3) {easy = !easy; if (easy) boop(20);}
        } else {    // ---- space between words detection
          delay(1);
          if (!ptime) p++;
          //        if (p>(4*ditPeriod)) {  // time for a space between words (3p after last char + 4p = 7p)
          pspace = (line[ci] == ' ') ? 3-1 : 5;   // adj. to tolarate speed variations, at least (-1 to make easier for new-bes)
          pspace =  (pspace + xt)*ditPeriod;  // xt normally =0 is for eXtra Time
          if (p > pspace) {  // if times up make their input a space
            cinput = ' ';    // if char expected it will err as too soon, if between words this must happen else err
          }
        }
        //  OK now?    if (cinput==' ' && line[ci-1]==' ') cinput='#';  // allow extra space at begining of words
        // If last or next char is puncuation allow for a space in time.
        //      if (cinput==' ' && (isPunct(line[ci-1]) || isPunct(line[ci]))) cinput='#';
        if (cinput == ' ' && (isPunct(line[ci]))) cinput = '#';
        if (ESC || btn4) return;
      }
      xt = 0;

      // -----------  did we get the char we expected
      testC = line[ci];
      if (testC == '[' || testC == '{')  testC = '(';
      if (testC == ']' || testC == '}')  testC = ')';
      if (cinput == ')' && testC == '(')  cinput == '(';
      if (easy && line[ci] == ' ' && cinput != ' ') {
        ci++;
        Serial.print('_');
      }
      if (cinput == toupper(testC)) { // ---- got it as expected
        ci++;
        if (cinput != ' ') chrs++;
        ptime = (cinput == ' ');
        Serial.print(cinput);
        if (line[ci] == ' ' || line[ci] == 0x0) { // update WPM tracking after each word
          wi++;
          wpm = ((1000.0 * 60.0) / (240.0 * ditPeriod / 5.0));
          if (wpmTkr==0) {wpmTkr=6*wpm; wpm_min=10*wpm; wpm_max=10*wpm;}
          wpmTkr = (5 * wpmTkr / 6 + wpm); // track & average over a practice round
          if ((10*wpm)<wpm_min) wpm_min=(wpm_min+10*wpm)/2; // track min. & max
          if ((10*wpm)>wpm_max) wpm_max=(wpm_max+10*wpm)/2;
        }
        if (lerr) xt = 3; // if they are just righting a wrong allow x-time to get going again
        lerr = false;

      } else {                       // ---- Not as expected
        if (easy && cinput == ' ') continue;
        buZZ(25); // wrong char || expected more time || 
        if (!lerr) {
          errs++;
          buZZ(75, 150);
        }
        lerr = true;
        ptime = true; // let them have time to regroup
        //      ptime=(line[ci]!=' ');  // if they gave wrong Letter-Symble let them have time to regroup
        if (line[ci] == ' ') {
          ci++;
          Serial.print("_");
          continue;
        }
      }
    }

    t = t1 - t0;
    wpm = ((1000.0 * 60.0) / (240.0 * ditPeriod / 5.0));
    // note: avg char code takes 12p, 4 char is 48p, 5* that is 240p, the std for 5WPM

    Serial.print("\n Trending ditdah Keying rate in WPM: "); Serial.print(wpmTkr / 6);
    //Serial.print("\nchrs: "); Serial.print(chrs); Serial.print("   t: "); Serial.println(t);

    wpm = (chrs / 4.5) / (t / (1000.0 * 60.0)); // adj. for the char distribution in our word databaase
    if (wpm < 1) wpm = 1;
    if (maxC>10) {
      Serial.print("    over all ~ Words/Minute: "); Serial.println(wpm);
      int vard= int(100* (((wpm_max-wpm_min)/10.0)/wpm) );
      if (maxC<30 && vard==0) vard=1;
      Serial.print("           WPM varied: "); Serial.print(vard);
      Serial.print("%   dah/dit Ratio (SB=3.0): "); Serial.println(float(avgDah)/avgDit);
    }
    else Serial.println();
    delay(750);
    if (errs > 0) {
      Serial.print("   Number of Incorrect characters: "); Serial.println(errs);
      buZZ(200 + 50 * errs);
      //    showLevel(errs);
      //    refreshWait(1000); clearDisp();
    } else {
      Serial.println("    No errors were detected.");
      if (maxC < 80) {
        maxC += 10;
        set_LEDs(0xFF, 3); delay(250); // congratulations Flash
        set_LEDs(0xFF, 0); delay(100);
        set_LEDs(0xFF, 3); delay(250);
        showLevel(maxC / 10);
        Serial.print("    Advance minimum line length to "); Serial.println(maxC);
        refreshWait(2000); clearDisp();
      }
    }
  }
}
